# -*- python -*-
# $RCSfile: labelledslider.py,v $
# $Revision: 1.20 $
# $Author: vrc $
# $Date: 2007-09-26 22:02:28 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

import gtk
from ooflib.common import utils
from ooflib.common import debug
from ooflib.common import guitop
from ooflib.common.IO.GUI import gtklogger

class LabelledSlider:
    def __init__(self, value=None, vmin=0, vmax=1, step=0.01, callback=None,
                 name=None):
        # "callback" is called when the user moves the slider.
        debug.mainthreadTest()
        self.gtk = gtk.HPaned()
        if value is None:
            value = vmin
        if name is not None:
            gtklogger.setWidgetName(self.gtk, name)
        self.adjustment = gtk.Adjustment(value=value,
                                         lower=vmin, upper=vmax,
                                         step_incr=step,
                                         page_incr=step)
        self.slider = gtk.HScale(self.adjustment)
        gtklogger.setWidgetName(self.slider, "slider")
        gtklogger.adoptGObject(self.adjustment, self.slider,
                              access_method=self.slider.get_adjustment)
        self.slider.set_size_request(100, -1)
        self.gtk.pack1(self.slider, resize=True, shrink=True)
        self.slider.set_draw_value(False)   # we'll display the value ourselves
        self.adjustmentsignal = gtklogger.connect(self.adjustment,
                                                  'value-changed',
                                                  self.text_from_slider)

        self.entry = gtk.Entry()
        gtklogger.setWidgetName(self.entry, "entry")
        self.gtk.pack2(self.entry, resize=True, shrink=True)

        # Make sure that the Entry is big enough to hold the min and
        # max values, or at least 8 digits.
        width = max(len(`vmin`), len(`vmax`), 8)
        self.entry.set_size_request(width*guitop.top().digitsize, -1)

        self.entrysignal = gtklogger.connect(self.entry, 'changed',
                                             self.entry_changed)

        # The initial call to set_value here should *not* call the
        # callback, since if this LabelledSlider is part of a
        # ParameterWidget, the rest of the widget might not be
        # constructed yet.  So here we temporarily set the callback
        # function to None.  Then the 'value-changed' signal can be
        # used to call self.text_from_slider without also calling the
        # callback function.  (The alternative is to postpone
        # connecting the 'value-changed' signal until after calling
        # set_value, but then we'd have to explicitly set the text in
        # self.entry.  That would be uglier.)
        self.callback = None
        self.set_value(value)
        # Now it's safe to set the callback.
        self.callback = callback

        gtklogger.connect(self.entry, 'activate', self.slider_from_text)
        gtklogger.connect(self.entry, 'focus-out-event', self.entry_lost_focus)
        

    def set_value(self, value):
        debug.mainthreadTest()
        if value < self.adjustment.lower:
            self.adjustment.set_value(self.adjustment.lower)
        elif value > self.adjustment.upper:
            self.adjustment.set_value(self.adjustment.upper)
        else:
            self.adjustment.set_value(value)
        self.adjustment.value_changed()
    def text_from_slider(self, obj):
        debug.mainthreadTest()
        val = self.slider.get_value()
        self.entrysignal.block()
        try:
            self.set_entry(val)
        finally:
            self.entrysignal.unblock()
        self.entry.set_position(0)
        if self.callback:
            self.callback(self, val)
    def slider_from_text(self, obj):    # callback for 'activate' from GtkEntry
        debug.mainthreadTest()
        try:
            v0 = self.get_value()
        except:
            # Illegal values will raise an exception, but they
            # may be incomplete entries.  So don't do anything
            # about it here.
            pass
        else:
            val = self.clip(v0)
            self.adjustmentsignal.block()
            try:
                self.adjustment.set_value(val)
            finally:
                self.adjustmentsignal.unblock()
            if self.callback:
                self.callback(self, val)
    def entry_lost_focus(self, obj, event):
        self.slider_from_text(obj)
    def entry_changed(self, obj):
        self.slider_from_text(obj)
    def clip(self, v):
        v = max(v, self.adjustment.lower)
        v = min(v, self.adjustment.upper)
        return v
    def parameterTableXRef(self, ptable, widgets):
        # Called after a LabelledSlider has been placed in a
        # ParameterTable.  All of the LabelledSliders in the table
        # should have their HPaned adjustments synchronized.
        self.syncsignals = [self.gtk.connect('notify', widget._syncothers)
                            for widget in widgets
                            if (isinstance(widget, LabelledSlider)
                                and widget is not self)]

    def _syncothers(self, pane, gparamspec):
        if gparamspec.name == 'position':
            pos = pane.get_property('position')
            for signal in self.syncsignals:
                self.gtk.handler_block(signal)
            self.gtk.set_position(pos)
            for signal in self.syncsignals:
                self.gtk.handler_unblock(signal)

class FloatLabelledSlider(LabelledSlider):
    def set_digits(self, digits):
        "Sets number of digits after the decimal place."
        debug.mainthreadTest()
        self.slider.set_digits(digits)
    def set_entry(self, val):
        debug.mainthreadTest()
        self.entry.set_text(("%-8g" % val).rstrip())
    def get_value(self):
        debug.mainthreadTest()
        return self.clip(utils.OOFeval(self.entry.get_text()))


class IntLabelledSlider(LabelledSlider):
    def __init__(self, *args, **kwargs):
        LabelledSlider.__init__(self, *args, **kwargs)
        self.slider.set_digits(0)
    def set_entry(self, val):
        debug.mainthreadTest()
        self.entry.set_text("%d" % int(val))
    def get_value(self):
        debug.mainthreadTest()
        return self.clip(int(utils.OOFeval(self.entry.get_text())))
